#include <alsa/asoundlib.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <fftw3.h>
#include <time.h>

// graphical spectra for debugging (enable in Makefile)
#ifdef ENABLE_GL
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <GL/gl.h>
#include <GL/glx.h>
#endif

 /*******************
   Helper functions
 ********************/


float min(float a, float b) {
  return a>b ? b : a;
}
float max(float a, float b) {
  return a>b ? a : b;
}

#ifdef ENABLE_GL
/*
 * Create an RGB, double-buffered window.
 * Return the window and context handles.
 */
static void
make_window( Display *dpy, const char *name,
             int x, int y, int width, int height,
             Window *winRet, GLXContext *ctxRet)
{
   int attrib[] = { GLX_RGBA,
                    GLX_RED_SIZE, 1,
                    GLX_GREEN_SIZE, 1,
                    GLX_BLUE_SIZE, 1,
                    GLX_DOUBLEBUFFER,
                    GLX_DEPTH_SIZE, 1,
                    None };
   int scrnum;
   XSetWindowAttributes attr;
   unsigned long mask;
   Window root;
   Window win;
   GLXContext ctx;
   XVisualInfo *visinfo;

   scrnum = DefaultScreen( dpy );
   root = RootWindow( dpy, scrnum );

   visinfo = glXChooseVisual( dpy, scrnum, attrib );
   if (!visinfo) {
      printf("Error: couldn't get an RGB, Double-buffered visual\n");
      exit(1);
   }

   /* window attributes */
   attr.background_pixel = 0;
   attr.border_pixel = 0;
   attr.colormap = XCreateColormap( dpy, root, visinfo->visual, AllocNone);
   attr.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask;
   mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;

   win = XCreateWindow( dpy, root, 0, 0, width, height,
                        0, visinfo->depth, InputOutput,
                        visinfo->visual, mask, &attr );

   /* set hints and properties */
   {
      XSizeHints sizehints;
      sizehints.x = x;
      sizehints.y = y;
      sizehints.width  = width;
      sizehints.height = height;
      sizehints.flags = USSize | USPosition;
      XSetNormalHints(dpy, win, &sizehints);
      XSetStandardProperties(dpy, win, name, name,
                              None, (char **)NULL, 0, &sizehints);
   }

   ctx = glXCreateContext( dpy, visinfo, NULL, True );
   if (!ctx) {
      printf("Error: glXCreateContext failed\n");
      exit(1);
   }

   XFree(visinfo);

   *winRet = win;
   *ctxRet = ctx;
}
#endif

class audioInput {
  public:
    char* data;
    timeval last_write;
      
      // Audio device data
    int bytes_per_frame;
    int read_frames, read_bytes;
    int rate; /* Sample rate */
    float dt;
    int exact_rate;   /* Sample rate returned by */
                      /* snd_pcm_hw_params_set_rate_near */ 
    int dir;          /* exact_rate == rate --> dir = 0 */
                      /* exact_rate < rate  --> dir = -1 */
                      /* exact_rate > rate  --> dir = 1 */
    int periods;       /* Number of periods */
    snd_pcm_uframes_t periodsize; /* Periodsize (bytes) */
    int size, exact_size;  // buffer size
      /* Handle for the PCM device */ 
    snd_pcm_t *pcm_handle;
    /* Playback stream */
    snd_pcm_stream_t stream;
    /* This structure contains information about    */
    /* the hardware and can be used to specify the  */      
    /* configuration to be used for the PCM stream. */ 
    snd_pcm_hw_params_t *hwparams;
    /* Name of the PCM device, like plughw:0,0          */
    /* The first number is the number of the soundcard, */
    /* the second number is the number of the device.   */
    char *pcm_name;

    audioInput() {
      gettimeofday(&last_write, NULL);
      bytes_per_frame = 2;
      read_frames = 1024;
      
      read_bytes = read_frames * bytes_per_frame;
      data = new char[read_bytes];
      
      rate = 44100;
      dt = 1.0 / (float)rate;
      periods = 2;
      periodsize = 8192;
      size = exact_size = (periodsize * periods)>>2;
      
      if( initDevice() < 0 ) {
        exit(1);
      }
      
    }
    ~audioInput() {
      snd_pcm_close (pcm_handle);
    }

    char* audioCapture() {
      int n;
      while((n = snd_pcm_readi(pcm_handle, data, read_frames)) < 0 ) {
        snd_pcm_prepare(pcm_handle);
        fprintf(stderr, "Dropped audio data.\n");
      }
      return data;
    }
    
    
    
    int initDevice() {
  
      stream = SND_PCM_STREAM_CAPTURE;
      
      /* Init pcm_name. Of course, later you */
      /* will make this configurable ;-)     */
      pcm_name = strdup("plughw:0,0");
      
      /* Allocate the snd_pcm_hw_params_t structure on the stack. */
      snd_pcm_hw_params_alloca(&hwparams);
      
      /* Open PCM. The last parameter of this function is the mode. */
      /* If this is set to 0, the standard mode is used. Possible   */
      /* other values are SND_PCM_NONBLOCK and SND_PCM_ASYNC.       */ 
      /* If SND_PCM_NONBLOCK is used, read / write access to the    */
      /* PCM device will return immediately. If SND_PCM_ASYNC is    */
      /* specified, SIGIO will be emitted whenever a period has     */
      /* been completely processed by the soundcard.                */
      if (snd_pcm_open(&pcm_handle, pcm_name, stream, 0) < 0) {
        fprintf(stderr, "Error opening PCM device %s\n", pcm_name);
        return(-1);
      }
      
      /* Init hwparams with full configuration space */
      if (snd_pcm_hw_params_any(pcm_handle, hwparams) < 0) {
        fprintf(stderr, "Can not configure this PCM device.\n");
        return(-1);
      }
      
        /* Set access type. This can be either    */
      /* SND_PCM_ACCESS_RW_INTERLEAVED or       */
      /* SND_PCM_ACCESS_RW_NONINTERLEAVED.      */
      /* There are also access types for MMAPed */
      /* access, but this is beyond the scope   */
      /* of this introduction.                  */
      if (snd_pcm_hw_params_set_access(pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED) < 0) {
        fprintf(stderr, "Error setting access.\n");
        return(-1);
      }
    
      /* Set sample format */
      if (snd_pcm_hw_params_set_format(pcm_handle, hwparams, SND_PCM_FORMAT_S16_LE) < 0) {
        fprintf(stderr, "Error setting format.\n");
        return(-1);
      }
  
      /* Set sample rate. If the exact rate is not supported */
      /* by the hardware, use nearest possible rate.         */ 
      exact_rate = rate;
      if (snd_pcm_hw_params_set_rate_near(pcm_handle, hwparams, (uint*)&exact_rate, 0) < 0) {
        fprintf(stderr, "Error setting rate.\n");
        return(-1);
      }
      if (rate != exact_rate) {
        fprintf(stderr, "The rate %d Hz is not supported by your hardware.\n \
                        ==> Using %d Hz instead.\n", rate, exact_rate);
      }
  
      /* Set number of channels */
      if (snd_pcm_hw_params_set_channels(pcm_handle, hwparams, 1) < 0) {
        fprintf(stderr, "Error setting channels.\n");
        return(-1);
      }
  
      /* Set number of periods. Periods used to be called fragments. */ 
      if (snd_pcm_hw_params_set_periods(pcm_handle, hwparams, periods, 0) < 0) {
        fprintf(stderr, "Error setting periods.\n");
        return(-1);
      }
      /* Set buffer size (in frames). The resulting latency is given by */
      /* latency = periodsize * periods / (rate * bytes_per_frame)     */
      if (snd_pcm_hw_params_set_buffer_size_near(pcm_handle, hwparams, (snd_pcm_uframes_t*)&exact_size) < 0) {
        fprintf(stderr, "Error setting buffersize.\n");
        return(-1);
      }
      if( size != exact_size ) {
        fprintf(stderr, "Buffer size %d is not supported, using %d instead.\n", size, exact_size);
      }
  
        /* Apply HW parameter settings to */
      /* PCM device and prepare device  */
      if (snd_pcm_hw_params(pcm_handle, hwparams) < 0) {
        fprintf(stderr, "Error setting HW params.\n");
        return(-1);
      }
      return 1;
    }

};


// Just a little hot debugging action.
void printFFT(int size, float* data) {
  int w = 150;
  int h = 40;
  int step = size / w;
  double max = 4.0;
  double min = -0.0;
  int c;
  for( int i=0; i<h; i++ ) {
    c = 0;
    double v = min + ((max-min)/h * (h-i));
    for( int j=0; j<w; j++ ) {
      if( fabs(data[c]) > v )
        printf("*");
      else
        printf(" ");
      c+=step;
    }
    printf("\n");
  }

}

#ifdef ENABLE_GL
void initGL() {
  glBlendFunc (GL_SRC_ALPHA, GL_DST_ALPHA);
  glEnable (GL_BLEND);
  glEnable (GL_ALPHA_TEST);
  glAlphaFunc (GL_ALWAYS, 0.5);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_ALPHA);
  glDisable(GL_LIGHTING);


  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glOrtho(0,800,-10,3,-1,1);
  
  glBegin(GL_LINES);
    glColor4f(1,1,1,0.1);
    glVertex2f(0,0);
    glVertex2f(800,0);
  glEnd();
}

void drawPlot(int n, float* fftData, float r, float g, float b) {
  initGL();
  n = 800;
  glBegin(GL_LINE_STRIP);
  glColor4f(r, g, b, 0.5);
  float dx = 800.0 / (float)n;
  float x = 0;
  for( int i=0; i<n; i++ ) {
    glVertex2f(x, log(fabs(fftData[i])));
    x += dx;
  }
  glEnd();
}
#endif

union byte {
  unsigned char uchar_val;
  char char_val;
};


int main(int argc, char** argv) {
  // Start GL window
  #ifdef ENABLE_GL
    Display *dpy;
    Window win;
    GLXContext ctx;
    char *dpyName = ":0";
    dpy = XOpenDisplay(dpyName);
    make_window(dpy, "Window!", 0, 0, 800, 600, &win, &ctx);
    XMapWindow(dpy, win);
    glXMakeCurrent(dpy, win, ctx);
  #endif

  audioInput* ai = new audioInput();
  float inv256 = 1.0 / 256.0;
  float inv256_2 = inv256*inv256;
  int chunkSize = 4096;  // process FFT in chinks this size. Approx. 10Hz
  int fftSize = chunkSize;  // size of FFT results
  int backStore = 48000;  // Store this many frames backward before discarding
  int readPtr = 0; // location to start writing audio data into buffer (long because it does not wrap)
  int fftPtr = 0; // location to start copying to FFT buffer
  
  short* audioBuffer = new short[backStore];
  
  // prepare FFT structures
  fftw_complex *fftData;
  fftw_plan fftPlan, fftPlanB;
  
  fftData = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * chunkSize);
  fftPlan = fftw_plan_dft_1d(chunkSize, fftData, fftData, FFTW_FORWARD, FFTW_ESTIMATE);
  fftPlanB = fftw_plan_dft_1d(chunkSize, fftData, fftData, FFTW_BACKWARD, FFTW_ESTIMATE);
  
  // data analysis structures
  int anSize = fftSize / 10; // We are only interested in analyzing the lower frequencies
  float* fftAvg = new float[fftSize];  // holds the background profile
  float* fftDev = new float[fftSize];  // holds the average deviation from background profile
  float* fftMag = new float[fftSize];  // holds the magnitude of the complex FFT values
  float* fftLimit = new float[fftSize];  // just for drawing pretty pictures
  
  
  int collectFrames = 100;  // collect averages for 100 iterations, then begin listening
  float triggerDev = 10.0;  // sound must be X times greater than average deviation to trigger event
  int linger = 1; // seconds to continue recording after sound has stopped
  
  char* fileName = new char[512];
  char* cmd = new char[1024];
  FILE* recPipe;
  long iter = 0;
  bool calibrating = true;
  bool recording = false;
  printf("Please keep quiet, calibrating...\n");
  
  // initialize the average arrays
  for( int i=0; i<fftSize; i++ ) {
    fftAvg[i] = 0.0;
    fftDev[i] = 1.0;
  }
  
  time_t lastHit;
  struct tm * timeinfo;
  
  while(1) {
      // Read in a block of data from the audio thread
    char* buf = ai->audioCapture();
    int size = ai->read_bytes;
    byte b;
    for( int i=0; i<size; i+= ai->bytes_per_frame ) {
      b.char_val = buf[i];
      float val = (float)buf[i+1]*inv256 + (float)b.uchar_val*inv256_2;
      
        // store in the fft buffer for analysis
      if( fftPtr < chunkSize ) {
        fftData[fftPtr][0] = val;
        fftData[fftPtr][1] = 0.0;
        fftPtr++;
      }
    }
    
      // Perform a FFT if the FFT buffer is full
    if( fftPtr >= chunkSize ) {
      #ifdef ENABLE_GL
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      #endif
      if( calibrating && ++iter > collectFrames ) {
        printf("Done calibrating.\n");
        calibrating = false;
      }
      fftPtr = 0;
      
      // Do the forward FFT
      fftw_execute(fftPlan);
      
      // Calculate the magnitude of all our complex wectors
      for( int i=0; i<fftSize; i++ ) {
        fftMag[i] = sqrt(fftData[i][0]*fftData[i][0] + fftData[i][1]*fftData[i][1]);
      }
      
      #ifdef ENABLE_GL
        drawPlot(fftSize, fftMag, 0.3, 0.3, 0.1);
        drawPlot(fftSize, fftAvg, 0.4, 0.8, 0.3);
        drawPlot(fftSize, fftDev, 0.6, 0.3, 0.3);
          
        for( int i=0; i<anSize; i++ ) {
          fftLimit[i] = fftAvg[i] + triggerDev * fftDev[i];
        }
        drawPlot(fftSize, fftLimit, 0.6, 0.0, 0.0);
      #endif
      
      if( calibrating ) {
        
        // averaging constants get smaller over time
        float ac = 1.0 / (float)iter;
        float dc = 1.0 / (float)iter;
        
        for( int i=0; i<fftSize; i++ ) {
          float fd = fftMag[i];
          float dev = fabs(fd - fftAvg[i]);
          fftDev[i] = (fftDev[i] * (1.0 - dc)) + (dev * dc);
          fftAvg[i] = (fftAvg[i] * (1.0 - ac)) + (fd * ac);
        }
      }
      else {
        time_t now;
        time( &now );
        
        // Stop recording if there has been no sound for a while
        if( recording && now - lastHit > linger ) {
          printf("Stopped recording.\n");
          recording = false;
          pclose(recPipe);
        }
        
        
        // Go over the low frequency part of the FFT and find anomalies, 
        //   start recording if needed.
        for( int i=0; i<anSize; i++ ) {
          float fd = fftMag[i];
          float dev = fabs(fd - fftAvg[i]);
          
          if( dev > fftDev[i] * triggerDev ) {
            lastHit = now;
            timeinfo = localtime ( &lastHit );
            i = anSize;
            if( ! recording ) {
              recording = true;
              sprintf(fileName, "%d.%02d.%02d-%02d.%02d.%02d", (timeinfo->tm_year + 1900), (timeinfo->tm_mon+1), timeinfo->tm_mday, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
              printf("Started recording to %s\n", fileName);
              
              // Open a pipe to flac
              sprintf(cmd, "/usr/bin/flac -s -o %s.flac --endian=little --sign=signed --channels=1 --bps=16 --sample-rate=44100 --force-raw-format -", fileName);
              recPipe = popen(cmd, "w");
//               recPipe = fopen(fileName, "w");
              
              // Go back and write the entire back buffer
              fwrite(audioBuffer+readPtr, sizeof(short), backStore-readPtr-1, recPipe);
              if( readPtr > 0 ) {
                fwrite(audioBuffer, sizeof(short), readPtr-1, recPipe);
              }
            }
          }
        }
      }  // end if(calibrating)
      
      
      // Process audio to remove background noise
      for( int i=0; i<fftSize; i++ ) {
      
        // First drop out the noise profile, curving the bottom end nicely
        float dev;
        float w = fftAvg[i] * 1.0;
//         float dev = fftMag[i] - fftAvg[i];
        if( fftMag[i] < 2.0*w ) {
          dev = fftMag[i] * fftMag[i] / (4.0 * w);
        }
        else {
          dev = fftMag[i] - w;
        }
          
        float scale = 1.0;
        scale *= dev / fftMag[i];

        // then remove anything that doesn't sound interesting
//         scale *= 1.0 / (1.0 + exp(exp(exp(1.0 - dev))));
        fftMag[i] *= scale;
        fftData[i][0] *= scale;
        fftData[i][1] *= scale;
      }
//      drawPlot(fftSize, fftMag, 0.3, 0.3, 0.7);
      fftw_execute(fftPlanB);
      float scale = 1.0 / (float)fftSize;
      
      // Write cleaned data into buffer and into pipe if recording
      char buf[3];
      for( int i=0; i<fftSize; i++ ) {
        float v = fftData[i][0] * scale;
        
          // store in the audio buffer in case we want to write it to disk later
        audioBuffer[readPtr] = (short)(v*256.0*256.0);
        if( recording ) {
          fwrite(audioBuffer+readPtr, sizeof(short), 1, recPipe);
        }
        readPtr++;
        if( readPtr > backStore )
          readPtr = 0;
      }
      
      #ifdef ENABLE_GL
        glFinish();
        glXSwapBuffers(dpy, win);
      #endif
      
    } // end if( fftPtr ... )
    
  }
  
  fftw_destroy_plan(fftPlan);
  fftw_free(fftData);
  return 0;
}

